@@ -5,12 +5,13 @@ from io import StringIO, BytesIO
 from lxml import etree
 from reportlab.lib.colors import HexColor, black
 from reportlab.lib.units import mm
-from reportlab.platypus import TableStyle, PageBreak, Spacer
+from reportlab.platypus import TableStyle, PageBreak, Spacer, Table
 
 from .enhanced_paragraph.enhanced_paragraph import EnhancedParagraph
 from .enhanced_paragraph.style import EnhancedParagraphStyle
 from .enhanced_table.data import EnhancedTableData
-from .enhanced_table.enhanced_tables import OVERFLOW_ROW, EnhancedTable, HEADER_FOOTER
+from .enhanced_table.enhanced_tables import OVERFLOW_ROW, EnhancedTable, HEADER_FOOTER, KEEP_TYPE_END, KEEP_TYPE_START, \
+    KEEP_TYPE_MIDDLE, KEEP_TYPE_SPAN, KEEP_TYPE_NA
 from .png_images import insert_image, insert_obj
 from .svglib.svglib import SvgRenderer
 from .utils import DocTemplate, get_page_size_from_string, intcomma_currency, ColumnWidthPercentage
@@ -19,7 +20,6 @@ from ..pagers.border import BorderPager
 
 
 class ReportXML(object):
-
     pager_types = {'borders': BorderPager}
     default_pager = BasePager
 
@@ -58,7 +58,7 @@ class ReportXML(object):
         (u'rsquo', u'’'),
     ]
 
-    def __init__(self, object_lookup=None, pager_kwargs=None):
+    def __init__(self, object_lookup=None, pager_kwargs=None, test_mode=False):
         self.styles = {}
         if object_lookup is not None:
             self.object_lookup = object_lookup
@@ -78,6 +78,7 @@ class ReportXML(object):
         self.border_top_continuation = 0
         self.border_bottom_continuation = 0
         self.page_style = None
+        self.test_mode = test_mode
 
         if pager_kwargs is None:
             self.pager_kwargs = {}
@@ -176,6 +177,7 @@ class ReportXML(object):
                               background_image_first=self.background_image_first,
                               background_image_remaining=self.background_image_remaining,
                               background_image_footer=self.background_image_footer,
+                              test_mode=self.test_mode,
                               **self.pager_kwargs)
 
     def canvasmaker(self, *args, **kwargs):
@@ -192,6 +194,7 @@ class ReportXML(object):
                               background_image_first=self.background_image_first,
                               background_image_remaining=self.background_image_remaining,
                               background_image_footer=self.background_image_footer,
+                              test_mode=self.test_mode,
                               **self.pager_kwargs,
                               **kwargs)
 
@@ -255,9 +258,13 @@ class ReportXML(object):
         row_heights = []
         keep_data = []
         other_styles = {}
+        held_cells = {}
 
         self.process_css_for_table(table, main_styles, other_styles)
 
+        if self.get_boolean_value(table.get('hidden')):
+            return None
+
         min_rows_top = int(table.get('min_rows_top', 0))
         min_rows_bottom = int(table.get('min_rows_bottom', 0))
 
@@ -276,6 +283,7 @@ class ReportXML(object):
         pos_y = table.get('pos_y')
 
         held_row_span = 1
+        hidden_columns = set()
 
         for element in table:
             if element.tag == 'tr':
@@ -290,51 +298,60 @@ class ReportXML(object):
                                                                    rows_variables=rows_variables,
                                                                    variables=variables,
                                                                    col_widths=col_widths,
-                                                                   table_width=table_width)
+                                                                   table_width=table_width,
+                                                                   hidden_columns=hidden_columns,
+                                                                   held_cells=held_cells)
 
                 if max_row_span > held_row_span:
                     held_row_span = max_row_span
+
                 if held_row_span > 1 or min_rows_top > 0:
-                    for _ in range(overflow_row_count + 1):
-                        keep_data.append(1)
-                        headers_index.append(current_header_index)
-                        footers_index.append(current_footer_index)
                     min_rows_top -= 1
+                    keep_type = KEEP_TYPE_SPAN
                 else:
-                    for _ in range(overflow_row_count + 1):
-                        keep_data.append(0)
-                        headers_index.append(current_header_index)
-                        footers_index.append(current_footer_index)
+                    keep_type = KEEP_TYPE_NA
+                for _ in range(overflow_row_count + 1):
+                    keep_data.append(keep_type)
+                    headers_index.append(current_header_index)
+                    footers_index.append(current_footer_index)
+
                 held_row_span -= 1
                 row_count += overflow_row_count
             elif element.tag == 'keep':
                 local_keep_data = []
+                keep_type = KEEP_TYPE_START
+                for child_element in element:
+                    if child_element.tag == 'tr':
+                        row_count += 1
+                        max_row_span, overflow_row_count = self.process_tr(tr_element=child_element,
+                                                                           data=main_data,
+                                                                           styles=main_styles,
+                                                                           other_table_styles=other_styles,
+                                                                           row_heights=row_heights,
+                                                                           row_count=row_count,
+                                                                           span=main_span,
+                                                                           rows_variables=rows_variables,
+                                                                           variables=variables,
+                                                                           col_widths=col_widths,
+                                                                           table_width=table_width,
+                                                                           hidden_columns=hidden_columns,
+                                                                           held_cells=held_cells)
+                        if max_row_span > held_row_span:
+                            held_row_span = max_row_span
+                        if held_row_span > 1 or min_rows_top > 0:
+                            min_rows_top -= 1
 
-                for tr in element:
-                    row_count += 1
-                    _, overflow_row_count = self.process_tr(tr_element=tr,
-                                                            data=main_data,
-                                                            styles=main_styles,
-                                                            other_table_styles=other_styles,
-                                                            row_heights=row_heights,
-                                                            row_count=row_count,
-                                                            span=main_span,
-                                                            rows_variables=rows_variables,
-                                                            variables=variables,
-                                                            col_widths=col_widths,
-                                                            table_width=table_width)
-                    if min_rows_top > 0:
-                        min_rows_top -= 1
+                        for index in range(overflow_row_count + 1):
+                            headers_index.append(current_header_index)
+                            footers_index.append(current_footer_index)
+                            local_keep_data.append(keep_type)
+                            keep_type = KEEP_TYPE_MIDDLE
+
+                        held_row_span -= 1
+                        row_count += overflow_row_count
 
-                    for index, _ in enumerate(range(overflow_row_count + 1)):
-                        headers_index.append(current_header_index)
-                        footers_index.append(current_footer_index)
-                        if index == 0:
-                            local_keep_data.append(2)
-                        else:
-                            local_keep_data.append(3)
                 if len(local_keep_data) > 0:
-                    local_keep_data[-1] = 3
+                    local_keep_data[-1] = KEEP_TYPE_END
                 keep_data += local_keep_data
 
             elif element.tag == 'no_headers':
@@ -345,8 +362,8 @@ class ReportXML(object):
 
             elif element.tag in ['header', 'footer']:
                 is_header = element.tag == 'header'
-
-                if is_header and int(element.get('output', "0")) == 1:
+                output = element.get('output', "0")
+                if is_header and self.get_boolean_value(output):
                     for tr in element:
                         row_count += 1
                         _, overflow_row_count = self.process_tr(tr_element=tr,
@@ -358,7 +375,9 @@ class ReportXML(object):
                                                                 span=main_span,
                                                                 rows_variables=rows_variables,
                                                                 variables=variables,
+                                                                held_cells=held_cells,
                                                                 col_widths=col_widths,
+                                                                hidden_columns=hidden_columns,
                                                                 table_width=table_width)
                 header_footer_span = {}
                 header_footer_data = []
@@ -379,7 +398,8 @@ class ReportXML(object):
                                     col_widths=col_widths,
                                     table_width=table_width,
                                     default_row_height=35,
-                                    is_header_or_footer=True)
+                                    is_header_or_footer=True,
+                                    hidden_columns=hidden_columns)
 
                 enhanced_table_data = EnhancedTableData(row_data=header_footer_data,
                                                         row_heights=header_footer_row_height,
@@ -396,7 +416,7 @@ class ReportXML(object):
         for x in range(1, min_rows_bottom + 1):
             if length - x < 0:
                 break
-            keep_data[length - x] = 1
+            keep_data[length - x] = KEEP_TYPE_SPAN
 
         h_align, v_align = self.get_alignment_details(main_styles)
 
@@ -458,8 +478,14 @@ class ReportXML(object):
         return x, y
 
     def process_tr(self, tr_element, data, styles, other_table_styles,
-                   row_heights, row_count, span, rows_variables, variables, col_widths, table_width,
-                   default_row_height=None, is_header_or_footer=False):
+                   row_heights, row_count, span, rows_variables, variables, col_widths, table_width, hidden_columns
+                   ,
+                   held_cells=None, default_row_height=None, is_header_or_footer=False):
+
+        max_row_span = 0
+        if self.get_boolean_value(tr_element.get('hidden')):
+            return max_row_span, 0
+
         row_data = []
         other_styles = {}
 
@@ -525,11 +551,33 @@ class ReportXML(object):
         rows_variables.append(row_variables)
         col_count = 0
         for index, td_element in enumerate(tr_element, 0):
-            if len(col_widths) < col_count + 1:
-                col_widths.append(None)
+
             if td_element.tag != 'td':
                 continue
 
+            name = td_element.get('name')
+            if name is not None and name in hidden_columns:
+                continue
+            if index in hidden_columns:
+                continue
+
+            hidden_column = td_element.get('hidden_column')
+            col_span = int(td_element.get('colspan', "1"))
+            row_span = int(td_element.get('rowspan', "1"))
+
+            if self.get_boolean_value(hidden_column):
+                if name is not None:
+                    hidden_columns.add(name)
+                else:
+                    for h_index in range(col_span):
+                        hidden_columns.add(index + h_index)
+                continue
+            if td_element.get('hidden'):
+                continue
+
+            if len(col_widths) < col_count + 1:
+                col_widths.append(None)
+
             # check that cell is not marked as a rowspan
             p_offset = span.get('%d-%d' % (row_count, col_count + offset), None)
             while p_offset is not None and p_offset > 0:
@@ -538,8 +586,6 @@ class ReportXML(object):
                 offset += p_offset
                 p_offset = span.get('%d-%d' % (row_count, col_count + offset), None)
 
-            col_span = int(td_element.get('colspan', "1"))
-            row_span = int(td_element.get('rowspan', "1"))
             if row_span > max_row_span:
                 max_row_span = row_span
 
@@ -635,7 +681,8 @@ class ReportXML(object):
                         row_variables[add_to_variable_name] = variables.get(add_to_variable_name)
                         currency_variable = add_to_variable_name + '__currency'
                         if currency_variable in variables:
-                            currency_value = symbol + intcomma_currency(float(variables.get(add_to_variable_name)) / 100.0)
+                            currency_value = symbol + intcomma_currency(
+                                float(variables.get(add_to_variable_name)) / 100.0)
 
                             row_variables[currency_variable] = currency_value
                             variables[currency_variable] = currency_value
@@ -648,10 +695,32 @@ class ReportXML(object):
             elif len(td_element) > 0:
                 xml = etree.tostring(td_element, pretty_print=True)
 
+                overflow_gt_height = td_element.get('overflow_gt_height')
                 overflow_gt_length = int(td_element.get('overflow_gt_length', 0))
                 style = self.process_css_for_table_paragraph_style(styles, row_count, col_count + offset)
 
-                if overflow_gt_length:
+                if overflow_gt_height is not None:
+
+                    display_object, overflow_row_count = self.process_overflow_height(
+                        td_element=td_element,
+                        index=index,
+                        overflow_gt_height=overflow_gt_height,
+                        xml=xml,
+                        style=style,
+                        col_widths=col_widths,
+                        table_width=table_width,
+                        col_span=col_span,
+                        styles=styles,
+                        offset=offset,
+                        col_count=col_count,
+                        row_count=row_count,
+                        row_span=row_span,
+                        row_data=row_data,
+                        overflow_rows=overflow_rows,
+                        rows_variables=rows_variables,
+                        held_cells=held_cells)
+
+                elif overflow_gt_length:
                     out_xml, style, overflow_row_count = self.overflow_cell(td_element=td_element,
                                                                             xml=xml,
                                                                             overflow_gt_length=overflow_gt_length,
@@ -673,6 +742,10 @@ class ReportXML(object):
                 display_object = td_element.text
                 if display_object is None:
                     display_object = ''
+
+            hold_cell = td_element.get('hold_cell')
+            if hold_cell is not None:
+                held_cells[hold_cell] = {'display_object': display_object, 'index': index, 'colspan': col_span}
             row_data.append(display_object)
 
             width = self.set_column_width(td_element.get('width'))
@@ -685,10 +758,10 @@ class ReportXML(object):
                 styles.append(('SPAN',
                                (col_count + offset, row_count),
                                (col_count + offset + col_span - 1, row_count + row_span - 1)))
-                for x in range(1, overflow_row_count):
+                for x in range(1, overflow_row_count+1):
                     styles.append(('SPAN',
                                    (col_count + offset, x + row_count),
-                                   (col_count + offset + col_span-1, x + row_count + row_span-1)))
+                                   (col_count + offset + col_span - 1, x + row_count + row_span - 1)))
 
                 if row_span > 1:
                     for r in range(1, row_span):
@@ -925,14 +998,17 @@ class ReportXML(object):
     @staticmethod
     def split_cell(xml, overflow_length):
         overflow_rows = []
+        held_working_tags = []
+        raw_parts = None
         offset = xml.find('>'.encode())
-        m = re.search("<br\s?/>".encode(), xml[overflow_length+offset:])
+        m = re.search("<br\s?/>".encode(), xml[overflow_length + offset:])
         if m:
             pos = m.start() + overflow_length + offset
             span = m.span()
             match_len = span[1] - span[0]
-            xml_parts = re.split("<br\s?/>".encode(), xml[pos+match_len:])
+            xml_parts = re.split("<br\s?/>".encode(), xml[pos + match_len:])
             next_xml = ''
+            raw_parts = []
             # this makes the xml valid again after it has been split
 
             for xml_part in [xml[:pos]] + xml_parts:
@@ -952,56 +1028,27 @@ class ReportXML(object):
                                 break
                     else:
                         working_tags.append((tag_name, tag))
+                raw_parts.append(xml_part)
 
+                held_working_tags.append(working_tags)
                 for tag in reversed(working_tags):
                     working_xml += ('</' + tag[0] + '>').encode()
                 overflow_rows.append(working_xml)
                 for tag in working_tags:
                     next_xml += tag[1]
 
-            return overflow_rows[0], overflow_rows[1:]
+            return overflow_rows[0], overflow_rows[1:], raw_parts, held_working_tags
 
-        return xml, overflow_rows
+        return xml, overflow_rows, raw_parts, held_working_tags
 
-    def overflow_cell(self, td_element, xml, overflow_gt_length, styles, style,  offset, col_count, row_count, col_span,
+    def overflow_cell(self, td_element, xml, overflow_gt_length, styles, style, offset, col_count, row_count, col_span,
                       row_span, row_data, overflow_rows, rows_variables):
         text = re.sub('<.*?>', '', str(xml))
         if len(text) > overflow_gt_length:
             overflow_length = int(td_element.get('overflow_length', 0))
-            xml, overflow_rows_xml = self.split_cell(xml, overflow_length)
+            xml, overflow_rows_xml, _, _ = self.split_cell(xml, overflow_length)
+
             if overflow_rows_xml:
-                for overflow_row_offset, row_xml in enumerate(overflow_rows_xml, 1):
-                    if row_xml == overflow_rows_xml[-1]:
-                        style_tag_name = 'bottom'
-                    else:
-                        style_tag_name = 'middle'
-
-                    self.process_css_for_table(tag=td_element,
-                                               styles=styles,
-                                               other_styles={},
-                                               start_col=col_count + offset,
-                                               start_row=row_count + overflow_row_offset,
-                                               end_col=col_count + offset + col_span - 1,
-                                               end_row=row_count + overflow_row_offset + row_span - 1)
-                    self.process_css_for_table(tag=td_element,
-                                               styles=styles,
-                                               other_styles={},
-                                               start_col=col_count + offset,
-                                               start_row=row_count + overflow_row_offset,
-                                               end_col=col_count + offset + col_span - 1,
-                                               end_row=row_count + overflow_row_offset + row_span - 1,
-                                               style_tag_name='overflow_%s_style' % style_tag_name,
-                                               class_tag_name='overflow_%s_class' % style_tag_name)
-                    style = self.process_css_for_table_paragraph_style(
-                        css=styles,
-                        row_count=row_count + overflow_row_offset,
-                        col_count=col_count + offset)
-                    overflow_object = EnhancedParagraph(row_xml, style, css_classes=self.styles)
-                    overflow_row = ['' for _ in range(len(row_data))]
-                    overflow_row.append(overflow_object)
-                    overflow_rows.append(overflow_row)
-                    rows_variables.append(rows_variables[-1])
-
                 self.process_css_for_table(tag=td_element,
                                            styles=styles,
                                            other_styles={},
@@ -1012,6 +1059,18 @@ class ReportXML(object):
                                            style_tag_name='overflow_top_style',
                                            class_tag_name='overflow_top_class')
 
+                self.process_over_flow_xml(overflow_rows_xml=overflow_rows_xml,
+                                           overflow_rows=overflow_rows,
+                                           rows_variables=rows_variables,
+                                           td_element=td_element,
+                                           styles=styles,
+                                           row_count=row_count,
+                                           col_count=col_count,
+                                           row_span=row_span,
+                                           col_span=col_span,
+                                           row_data=row_data,
+                                           offset=offset)
+
                 style = self.process_css_for_table_paragraph_style(styles, row_count, col_count + offset)
         return xml, style, len(overflow_rows)
 
@@ -1021,3 +1080,128 @@ class ReportXML(object):
             entity_string += u'<!ENTITY %s \'%s\'>' % (entity[0], entity[1])
         xml = """<!DOCTYPE root SYSTEM "print_engine" [%s]>""" % entity_string
         return xml
+
+    @staticmethod
+    def get_overflow_overflow_gt_height(raw_overflow_gt_height, held_cells, widths):
+        if raw_overflow_gt_height.startswith('#'):
+            held_cell_data = held_cells.get(raw_overflow_gt_height[1:])
+            display_object = held_cell_data['display_object']
+            index = held_cell_data['index']
+            col_span = held_cell_data['colspan']
+            avail_width = sum(widths[index:index + col_span])
+
+            if isinstance(display_object, EnhancedParagraph):
+                return display_object.calc_text_height(avail_width=avail_width)
+            elif isinstance(display_object, Table):
+                _, h = display_object.wrap(0, 0)
+                return h
+            else:
+                return 5 * mm
+        else:
+            return float(raw_overflow_gt_height) * mm
+
+    def process_overflow_height(self, td_element, index, overflow_gt_height, xml, style,
+                                col_widths, table_width, col_span,
+                                styles, offset, col_count, row_count, row_span, row_data, overflow_rows,
+                                rows_variables, held_cells):
+        widths = self.process_column_widths(col_widths, table_width)
+        overflow_gt_height = self.get_overflow_overflow_gt_height(raw_overflow_gt_height=overflow_gt_height,
+                                                                  held_cells=held_cells,
+                                                                  widths=widths)
+
+        avail_width = sum(widths[index:index+col_span])
+        display_object = EnhancedParagraph(xml, style, css_classes=self.styles)
+        height = display_object.calc_text_height(avail_width=avail_width)
+
+        if height <= overflow_gt_height:
+            return display_object, 0
+
+        xml, overflow_rows_xml, raw_parts, held_working_tags = self.split_cell(xml=xml, overflow_length=0)
+
+        working_height = 0
+        working_xml = b''
+        found_break = False
+        overflow_row_offset = 0
+        for overflow_row_offset, (row_xml, raw_part, working_tags) in\
+                enumerate(zip([xml] + overflow_rows_xml, raw_parts, held_working_tags), 1):
+            working_xml += raw_part + '<br/>'.encode()
+            if raw_part == b'\n':
+                row_xml = row_xml.replace(b'\n', b'M')
+            p = EnhancedParagraph(row_xml, style, css_classes=self.styles)
+            height = p.calc_text_height(avail_width=avail_width)
+            working_height += height
+            if working_height > overflow_gt_height:
+                for tag in reversed(working_tags):
+                    working_xml += ('</' + tag[0] + '>').encode()
+                    found_break = True
+                break
+        overflow_len = len(overflow_rows_xml[overflow_row_offset-1:])
+        if not found_break or overflow_len == 0:
+            return display_object, 0
+
+        display_object = EnhancedParagraph(working_xml, style, css_classes=self.styles)
+
+        self.process_css_for_table(tag=td_element,
+                                   styles=styles,
+                                   other_styles={},
+                                   start_col=col_count + offset,
+                                   start_row=row_count,
+                                   end_col=col_count + offset + col_span - 1,
+                                   end_row=row_count + row_span - 1,
+                                   style_tag_name='overflow_top_style',
+                                   class_tag_name='overflow_top_class')
+
+        over_flow_len = self.process_over_flow_xml(overflow_rows_xml=overflow_rows_xml[overflow_row_offset - 1:],
+                                                   overflow_rows=overflow_rows,
+                                                   rows_variables=rows_variables,
+                                                   td_element=td_element,
+                                                   styles=styles,
+                                                   row_count=row_count,
+                                                   col_count=col_count,
+                                                   row_span=row_span,
+                                                   col_span=col_span,
+                                                   row_data=row_data,
+                                                   offset=offset)
+        return display_object, over_flow_len
+
+    def process_over_flow_xml(self, overflow_rows_xml, overflow_rows, rows_variables,
+                              td_element, styles, row_count, col_count, row_span, col_span, row_data, offset):
+
+        for overflow_row_offset, row_xml in enumerate(overflow_rows_xml, 1):
+            if row_xml == overflow_rows_xml[-1]:
+                style_tag_name = 'bottom'
+            else:
+                style_tag_name = 'middle'
+
+            self.process_css_for_table(tag=td_element,
+                                       styles=styles,
+                                       other_styles={},
+                                       start_col=col_count + offset,
+                                       start_row=row_count + overflow_row_offset,
+                                       end_col=col_count + offset + col_span - 1,
+                                       end_row=row_count + overflow_row_offset + row_span - 1)
+            self.process_css_for_table(tag=td_element,
+                                       styles=styles,
+                                       other_styles={},
+                                       start_col=col_count + offset,
+                                       start_row=row_count + overflow_row_offset,
+                                       end_col=col_count + offset + col_span - 1,
+                                       end_row=row_count + overflow_row_offset + row_span - 1,
+                                       style_tag_name='overflow_%s_style' % style_tag_name,
+                                       class_tag_name='overflow_%s_class' % style_tag_name)
+            style = self.process_css_for_table_paragraph_style(
+                css=styles,
+                row_count=row_count + overflow_row_offset,
+                col_count=col_count + offset)
+            overflow_object = EnhancedParagraph(row_xml, style, css_classes=self.styles)
+            overflow_row = ['' for _ in range(len(row_data))]
+            # noinspection PyTypeChecker
+            overflow_row.append(overflow_object)
+            overflow_rows.append(overflow_row)
+            rows_variables.append(rows_variables[-1])
+
+        return len(overflow_rows_xml)
+
+    # noinspection PyMethodMayBeStatic
+    def get_boolean_value(self, value):
+        return value in ['1', 'True', 'yes']
